'From Squeak3.2gamma of 15 January 2002 [latest update: #4743] on 25 July 2002 at 11:25:24 pm'!
Object subclass: #ObjectGraphTraversalStrategy
	instanceVariableNames: 'substitutionBlock substitutes valueWhenBlock traverseWhenBlock traverseClassesWhenBlock '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Object traversal'!

!ObjectGraphTraversalStrategy commentStamp: '<historical>' prior: 0!
Use this class when you need highly customized object graph traversal
behavior.

The primary function provided is dynamic object-substitution while the
graph is being walked.  Substitution does not mutate any objects in anyway,
it merely values your block using the substitute instead of the original.
Only one substitute will be created for each object in the graph.

Instance substitution is supported, usually used to substitute globals such
as Smalltalk and Transcript.  Additionally, general substitution behavior is
provided as a backup when specific instances cannot be provided.

Note that substitution is the very first thing that happens prior to any other
customizable functions provided by this class.  For example, the traverseWhen:
block is evaluated against the substitute, not the original.

traverseClasses is provided to include the implicit pointer to every objects class.
!


!Object methodsFor: 'actions'!
maGraphDo: threeArgBlock

	self
		maGraphWhere: [ :path :parent :index | true ]
		do: threeArgBlock
! !

!Object methodsFor: 'actions' stamp: 'cmm 1/24/2002 14:20'!
maGraphDo: threeArgBlock using: anObjectGraphTraversalStrategy
	"Walk the receiver and all of its variables valuing threeArgBlock.
	The first argument is an OrderedCollection of the objects leading to the current
	object, including the current object, the second argument is the parent
	of the current object (second to last in the path).  The third argument is
	the index or key object required to send instVarAt: to access the current
	object (except for Dictionaries, in which the key may not be an Integer.
	All arguments are in their substituted form."

	self
		maGraphDo: threeArgBlock
		using: anObjectGraphTraversalStrategy
		path: (OrderedCollection with: self)
		with: IdentitySet new
		with: Set new
! !

!Object methodsFor: 'actions'!
maGraphWhere: threeArgConditionBlock do: threeArgBlock

	| strategy |

	strategy := ObjectGraphTraversalStrategy new traverseWhen: threeArgConditionBlock.
	self
		maGraphDo: threeArgBlock
		using: strategy
! !

!Object methodsFor: 'private actions'!
maGraphDo: aBlock using: aObjectTraversalStrategy path: path with: alreadyVisitedSet with: largeIntegersSet

	1
		to: self maInstSize
		do:
			[ :varIndex |
			self
				maValueGraphNode: (self instVarAt: varIndex)
				index: varIndex
				using: aObjectTraversalStrategy
				with: aBlock
				path: path
				with: alreadyVisitedSet
				with: largeIntegersSet ]
! !

!Object methodsFor: 'private actions'!
maValueGraphNode: anObject index: anInteger using: aObjectTraversalStrategy with: aBlock path: path with: alreadyVisitedSet with: largeIntegersSet

	| stepInto substitute |

	anObject == nil ifTrue: [ ^anObject ].  "We sha'n't be checking the substitutes for nil.."
	anObject maIsMinimalObject ifTrue: [ ^anObject ].  "Never traverse or value a proxy.."
	substitute := aObjectTraversalStrategy substituteFor: anObject.
	(alreadyVisitedSet includes: substitute) ifTrue: [ ^substitute ].
	path add: substitute.
	stepInto :=
		aObjectTraversalStrategy
			shouldTraverseIntoLastOf: path
			from: self
			index: anInteger.
	(aObjectTraversalStrategy
		shouldValueLastOf: path
		from: self
		index: anInteger)
			ifTrue:
				[ alreadyVisitedSet add: substitute.  "IMPORTANT, this needs to be here and not above.."
				aBlock
					value: path
					value: self
					value: anInteger ].
	stepInto
		ifTrue:
			[ substitute
				maGraphDo: aBlock
				using: aObjectTraversalStrategy
				path: path
				with: alreadyVisitedSet
				with: largeIntegersSet ].
	path removeLast.
	(aObjectTraversalStrategy shouldTraverseClassOf: self)
		ifTrue:
			[ self
				maValueGraphNode: self class
				index: 0
				using: aObjectTraversalStrategy
				with: aBlock
				path: path
				with: alreadyVisitedSet
				with: largeIntegersSet ]
! !


!Collection methodsFor: 'private actions'!
maGraphDo: aBlock using: aObjectTraversalStrategy path: path with: alreadyVisitedSet with: largeIntegersSet

	| varIndex |

	varIndex := 1.
	self do:
		[ :each |
		self
			maValueGraphNode: each
			index: varIndex
			using: aObjectTraversalStrategy
			with: aBlock
			path: path
			with: alreadyVisitedSet
			with: largeIntegersSet.
		varIndex := varIndex + 1 ]
! !


!Dictionary methodsFor: 'private actions'!
maGraphDo: aBlock using: aObjectTraversalStrategy path: path with: alreadyVisitedSet with: largeIntegersSet

	| varIndex |

	varIndex := 1.
	self keysAndValuesDo:
		[ :eachKey :eachValue |
		[ :obj |
			self
				maValueGraphNode: obj
				index: varIndex
				using: aObjectTraversalStrategy
				with: aBlock
				path: path
				with: alreadyVisitedSet
				with: largeIntegersSet.
			varIndex := varIndex + 1 ]
				value: eachKey;
				value: eachValue ]
! !


!Interval methodsFor: 'private actions'!
maGraphDo: aBlock using: aObjectTraversalStrategy path: path with: alreadyVisitedSet with: largeIntegersSet

	1
		to: self maInstSize
		do:
			[ :varIndex | | stepInto |
			self
				maValueGraphNode: (self instVarAt: varIndex)
				index: varIndex
				using: aObjectTraversalStrategy
				with: aBlock
				path: path
				with: alreadyVisitedSet
				with: largeIntegersSet ]
! !


!ObjectGraphTraversalStrategy methodsFor: 'private initializing'!
initialize

	substitutionBlock := [ :obj | obj ].
	substitutes := IdentityDictionary new.
	valueWhenBlock := [ :path :parent :instVarIndex | true ].
	traverseWhenBlock := [ :path :parent :instVarIndex | true ].
	traverseClassesWhenBlock := [ :obj | false ]
! !

!ObjectGraphTraversalStrategy methodsFor: 'testing'!
shouldTraverseClassOf: anObject

	^traverseClassesWhenBlock value: anObject
! !

!ObjectGraphTraversalStrategy methodsFor: 'testing'!
shouldTraverseIntoLastOf: path from: parent index: anInteger

	^traverseWhenBlock
		value: path
		value: parent
		value: anInteger
! !

!ObjectGraphTraversalStrategy methodsFor: 'testing'!
shouldValueLastOf: anOrderedCollection from: parentObject index: anInteger

	^valueWhenBlock
		value: anOrderedCollection
		value: parentObject
		value: anInteger
! !

!ObjectGraphTraversalStrategy methodsFor: 'building'!
substitute: sourceObject with: substituteObject
	"Use this method to build an instance-based substitution dictionary
	used to avoid walking into certain objects (for example, Smalltalk).
	If anObject qualifies for direct substitution via this dictionary, it will
	not employ the general substitution specified by
	substituteNonSpecifiedNodesWith:"

	substitutes
		at: sourceObject
		put: substituteObject
! !

!ObjectGraphTraversalStrategy methodsFor: 'building'!
substituteNonSpecifiedNodesUsing: oneArgBlock
	"Allow each object encountered to be substituted with another object.
	Not only will the do: block operate on the substitute, the traverseWhen
	condition will be based on the current substitute.  Note that substitution
	does not mutate any objects, it only affects the behavior of the traversal."

	substitutionBlock := oneArgBlock
! !

!ObjectGraphTraversalStrategy methodsFor: 'building'!
traverseClassesWhen: oneArgBlock
	"oneArgBlock must answer a boolean indicating whether the class of its
	argument should be traversed.  oneArgBlock takes the object whose class
	we are asking whether to traverse."

	traverseClassesWhenBlock := oneArgBlock
! !

!ObjectGraphTraversalStrategy methodsFor: 'building'!
traverseWhen: threeArgBlock
	"The default traversal behavior is to traverse everything.  threeArgBlock
	may be specified to restrict the default traversal (stepping into an objects
	instance variables).  The first argument is an OrderedCollection of the
	path of objects that lead to the current node, the second is the parent
	of the current node (and second to last in the path), the last is the integer
	index of the instVar of the current node."

	traverseWhenBlock := threeArgBlock
! !

!ObjectGraphTraversalStrategy methodsFor: 'building'!
valueWhen: threeArgBlock
	"The default behavior is to value every node once (the framework automatically
	keeps track of where it has already been).   threeArgBlock takes the same arguments
	as the other parameters of this class."

	valueWhenBlock := threeArgBlock
! !

!ObjectGraphTraversalStrategy methodsFor: 'accessing'!
substituteFor: anObject

	^substitutes
		at: anObject
		ifAbsent: [ substitutionBlock value: anObject ]
! !


!ObjectGraphTraversalStrategy class methodsFor: 'creation'!
new

	^super new initialize
! !

